from hec.script import *
from hec.hecmath import * 
from wcds.script import *

import hec, rma, wcds
__doc__ = '''
This module allows Jython scripts to easily manipulate the CAVI by changing
modules, opening watersheds, and running forecasts.

The ForecastDSSFile class provides all the functionality of the DSSFile class, with the added 
features of having the time window initially set the RunTimeWindow of the forecast for which it 
is for, and for translating DSS path names with blank E- and F-parts into full path names based
on the forecast time interval, model, alternative and trial.

The API is as follows :
	
	==[Functions]==============================================================================
	
	setModule(moduleName)
		This function sets the CAVI to the specified module and returns name of the name
		of the module.  Valid modules are:
			
			"Data Acquisition"
			"Data Visualization"
			"Model Interface"
			"Watershed Setup"
			
	openWaterhshed(watershedName)
		This function opens the specified watershed and returns the name of the opened
		watershed.
		
	==[Forecast constructor]===================================================================
	
	Forecast(name, alternatives [, startTime, endTime]) 
		Creates a new Forecast object that represents a new or existing CAVI forecast with
		the specified name and alternatives.  The name is a string of the format 
		YYYY.MM.DD-HHmm and the alternatives is a sequence of valid alternative names.  If 
		all alternatives are to be used, then an empty sequence can be supplied.  If 
		startTime and endTime are supplied then a new forecast is created with the lookback
		time equal to startTime, the end time equal to endTime, and the forecast time 
		specified in the name.  If startTime and endTime are not supplied, then an existing
		forecast is opened.
		
	==[Forecast Object methods]================================================================
	
	getAlternativeKey(name)
		Returns the forecast alternative key associated with the specified name.  If a
		forecast alternative key is specified instead of a name, the same key is returned.
		If no key is associated with the specified name, None is returned.
		
	getAlternativeName(key)
		Returns the forecast alternative name associated with the specified key.  If no
		forecast name is associated with the specified key, None is returned.
		
	getAlternativeKeys()
		Returns a list of all valid forecast alternative keys.
		
	getAlternativeNames()
		Returns a list of all valid forecast alternative names.
		
	getTimeStepString()
		Returns a string representation of the forecast time interval.
		
	getRunTimeWindow()
		Returns a RunTimeWindow object for the forecast.
		
	getLookbackTimeString()
		Returns a string representation of the lookback (start) time of the forecast.
	
	getStartTimeString()
		Returns a string representation of the lookback (start) time of the forecast.
	
	getEndTimeString()
		Returns a string representation of the end time of the forecast.
		
	getForecastTimeString()
		Returns a string representation of the forecast time.
		
	getForecastDSSFileName()
		Returns the name of the forecast DSS file.
		
	getDSSFile()
		Returns a ForecastDSSFile object for the forecast.
		
	compute()
		Executes the forecast.
	
	==[ForecastDSSFile object methods]=========================================================
	
	setTimeWindow([startTime, endTime])
		Sets the forecast dss file's default time window to start and end at the specified
		times.  If startTime and endTime are not supplied then the default time window is
		reset to the default of the forecast lookback time and end time.
		
	setModel(model)
		Sets the model to use when generating path name F-parts.  The model may be supplied
		as and integer ordinal position in the model sequence or as one of the following
		strings: "MFP", "HMS", "ResSim", "RAS", "FIA".
		
	setAlternative(alternative)
		Sets the alternative to use when generating path name F-parts.  The supplied altern-
		ative may be an alternative name or key.
		
	setTrial(trial)
		Sets the trial number to use when generating path name F-parts.  If not set, the 
		trial defaults to 0.
		
	read(pathName)
		Returns the HecMath object read from the forecast DSS file with the specified path name.
		
	write(HecMath)
		Writes the specified HecMath object to the forecast DSS file and returns the status.
		
	close()
		Closes the forecast DSS file.
	

Mike Perryman
USACE HEC
09 Sep 2003
'''
#=========================================================================
class Forecast :
	__doc__ = '''
	Wrapper for wcds.cavi.model.Forecast that provides some additional
	functionality for CAVI scripting.
	'''
	def __init__(self, name, alternatives, start=None, end=None) :
		self._name_ = name
		self._alternatives_ = alternatives
		self._alternativeNames_ = None
		self._alternativeKeys_ = None
		if start :
			self._forecast_ = setupNewForecastRun(
				name,
				alternatives,
				start,
				end)
		else :
			self._forecast_ = setupExistingForecastRun(
				name, 
				alternatives)
			
	def getAlternativeKey(self, name) :
		if self._alternativeKeys_ == None :
			dict = getAlternativeNamesAndKeys(self._forecast_)
			self._alternativeKeys_= dict["key"]
			self._alternativeNames_ = dict["name"]
		try :
			return self._alternativeKeys_[name]
		except NameError :
			return None
			
	def getAlternativeName(self, key) :
		if self._alternativeNames_ == None :
			dict = getAlternativeNamesAndKeys(self._forecast_)
			self._alternativeKeys_= dict["key"]
			self._alternativeNames_ = dict["name"]
		try :
			return self._alternativeNames_[key]
		except NameError :
			return None
			
	def getDSSFile(self) :
		"A docstring"
		return ForecastDSSFile(self)
			
	def compute(self) :
		return self._forecast_.compute()
		
	def getAlternativeKeys(self) :
		return self._forecast_.getAlternatives()
			
	def getAlternativeNames(self) :
		return vectorToList(self._forecast_.getForecastRuns())
			
	def getForecastDSSFilename(self) :
		return self._forecast_.getForecastDSSFilename()
			
	def getTimeStepString(self) :
		return self._forecast_.getTimeStepString()
			
	def getRunTimeWindow(self) :
		return self._forecast_.getRunTimeWindow()
			
	def getLookbackTimeString(self) :
		return self._forecast_.getRunTimeWindow().getLookbackTimeString()
			
	def getStartTimeString(self) :
		return self._forecast_.getRunTimeWindow().getStartTimeString()
			
	def getForecastTimeString(self) :
		return self._forecast_.getRunTimeWindow().getStartTimeString()
			
	def getEndTimeString(self) :
		return self._forecast_.getRunTimeWindow().getEndTimeString()
			
#=========================================================================
class ForecastDSSFile :
	__doc__ = '''
	Wrapper for DSSFile specific to forecasts, sets default time window
	to the forecast time window, and builds the e- and f-parts of DSS
	pathnames from the forecast, alternative, and trial settings.
	'''
	def __init__(self, forecast) :
		self._forecast_ = forecast
		self._dssFile_ = DSS.open(forecast.getForecastDSSFilename())
		self._ePart_ = forecast.getTimeStepString()
		self._alternativeKey_ = None
		self._trialNumber_ = 0
		self._modelName_ = None
		self.setTimeWindow()
		
	def setTimeWindow(self, start=None, end=None) :
		if start :
			self._dssFile_.setTimeWindow(start, end)
		else :
			self._dssFile_.setTimeWindow(
				self._forecast_.getLookbackTimeString(), 
				self._forecast_.getEndTimeString())
				
	def getPathname(self, pathname, model=None, alternative=None, trial=None) :
		i = {"A" : 1, "B" : 2, "C" : 3, "D" : 4, "E" : 5, "F" : 6}
		pathParts = pathname.split("/")
		if not pathParts[i["E"]] : pathParts[i["E"]] = self._ePart_
		if not pathParts[i["F"]] :
			if not model : 
				model = self._modelName_
			if not alternative : 
				alternative = self._alternativeKey_
			else :
				alternative = self._forecast_.getAlternativeKey(alternative)
			if not trial :
				trial = self._trialNumber_
			pathParts[i["F"]] = makeFPart(model, alternative, trial)
		return "/".join(pathParts)
		
	def setAlternative(self, alternative) :
		self._alternativeKey_ = self._forecast_.getAlternativeKey(alternative)
		
	def setModel(self, model) :
		self._modelName_ = model
		
	def setTrial(self, trial) :
		self._trialNumber_ = trial
		
	def read(self, pathName) :
		return self._dssFile_.read(pathName)
		
	def write(self, data) :
		return self._dssFile_.write(data)
		
	def close(self) :
		self._dssFile_.close()
		
#=========================================================================
def openWatershed(watershedName) :
	'''
	Opens a specified watershed in the CAVI.
	'''
	if CAVI.isWatershedOpened() :
		#------------------------------------------------------#
		# some watershed is open - see if it's the one we want #
		#------------------------------------------------------#
		if CAVI.getWatershedName() != watershedName :   
			#---------------------------------#
			# switch to the desired watershed #
			#---------------------------------#
			CAVI.openWatershed(watershedName)
	else :
		#---------------------------------------------#
		# no watershed is open, open one from scratch #
		#---------------------------------------------#
		CAVI.openWatershed(rma.util.RMAIO.separator.join( \
			[hec.client.ClientApp.app().getAppStartDir(), \
			 "watershed", \
			 "base", \
			 watershedName, \
			 watershedName + ".wksp"]))

	#---------------------------------#
	# verify that we were successfull #
	#---------------------------------#
	openWatershed = CAVI.getWatershedName()
	if openWatershed != watershedName :
		raise Exception, "Unable to open watershed %s" % watershedName
		
	return openWatershed

#=========================================================================
def setModule(moduleName) :
	'''
	Sets the CAVI to the specified module, and returns the module.
	'''
	if `CAVI.getCurrentModule()` != moduleName : 
		CAVI.selectModule(moduleName)
	
	currentModule = CAVI.getCurrentModule()

	if `currentModule` != moduleName :
		raise Exception, "Unable to switch to %s module." % moduleName

	return currentModule

#=========================================================================
def setupNewForecastRun(
	forecastName,       # yyyy.mm.dd-hhmm
	alternativeNames,   # [] for all
	lookbackTime,       # string or HecTime
	endTime) :          # string or HecTime
	'''
	Sets up a new forecast to be run from a script via the
	compute() method.
	'''
	#-----------------------#                            
	# set the forecast time #
	#-----------------------#                            
	forecastTime = hec.heclib.util.HecTime()
	try :		    
		dateStr, timeStr = forecastName.split("-")
		yr, mon, day = dateStr.split(".")
		dateStr = "/".join([mon, day, yr])
		timeStr =  dateStr + " " + timeStr
		forecastTime.set(timeStr)
	except Exception, e :
		message = "Invalid forecast name"
		try     : message += " : " + `e`
		finally : message += "."
		raise Exception, message
		
	#-----------------------#
	# set the lookback time #
	#-----------------------#
	if type(lookbackTime) == type("") :
		timeStr = lookbackTime.strip()
		try :
			lookbackTime = hec.heclib.util.HecTime(timeStr)
			if not lookbackTime.isDefined() : raise Exception, "Undefined"
		except :
			if not timeStr.startswith("-") : timeStr = "-" + timeStr
			dummyTime = hec.heclib.util.HecTime()
			lookbackTime = hec.heclib.util.HecTime(forecastTime)
			if hec.heclib.util.HecTime.getTimeWindow(timeStr, lookbackTime, dummyTime) :
			  raise Exception, "Cannot parse lookback time : %s" % timeStr
		
	#------------------#
	# set the end time #
	#------------------#
	if type(endTime) == type("") :
		timeStr = endTime.strip()
		try :
			endTime = hec.heclib.util.HecTime(timeStr)
			if not endTime.isDefined() : raise Exception, "Undefined"
		except :
			if not timeStr.startswith("+") : timeStr = "+" + timeStr
			endTime = hec.heclib.util.HecTime(forecastTime)
			dummyTime = hec.heclib.util.HecTime()
			if hec.heclib.util.HecTime.getTimeWindow(timeStr, endTime, dummyTime) :
			  raise Exception, "Cannot parse lookback time : %s" % timeStr
	
	#---------------------#  
	# create the forecast #
	#---------------------#  
	try :		                                                                     
		forecast = wcds.script.Forecast.createForecast(forecastTime.toString(4), 
			                                 lookbackTime.toString(4), 
			                                 endTime.toString(4))    
		if not forecast : 
			raise Exception, "Cannot create forecast"
		forecast.setAlternatives(wcds.script.Forecast.getAltList())
	except Exception, e :
		try    : message = `e`
		except : message = "Cannot create forecast"
		raise Exception, message

	setupForecastRun(forecast, alternativeNames)

	return forecast

#=========================================================================
def setupExistingForecastRun(forecastName, alternativeNames) :
	'''
	Sets up an existing forecast to be run from a script via the
	compute() method.
	'''
	try :
		forecast = wcds.script.Forecast.openForecast(forecastName)
	except Exception, e :
		try    : message = `e`
		except : message = 'Forecast "%s" not available.' % forecastName
		raise Exception, message
	
	setupForecastRun(forecast, alternativeNames)
	
	return forecast

#=========================================================================
def setupForecastRun(forecast, alternativeNames) :
	'''
	Common code for setting up new and existing forecast runs.
	'''
	#-------------------#
	# load the forecast #
	#-------------------#
	if not forecast.isLoaded() :
		forecast.loadWorkspace(
			  hec.client.ClientApp.app().getAppServer().getBindingSites()[0],
			  hec.client.ClientApp.app().getAppStartDir())

	if not forecast.isLoaded() :
		raise Exception, 'Cannot load forecast "%s".' % forecastName

	module = CAVI.getCurrentModule()
	module.setForecast(forecast)

	#------------------------------#
	# get the desired alternatives #
	#------------------------------#
	alternativesToRun = []
	alternatives = forecast.getForecastRuns()
	for alternative in alternatives :
		if not alternativeNames or (`alternative` in alternativeNames) :
			alternative.setActiveRun(Constants.TRUE)
			alternativesToRun.append(alternative)
		else :
			alternative.setActiveRun(Constants.FALSE)

	if not alternativesToRun :
		raise Exception, "No alternatives to run!"
	
	hec.client.ClientApp.app().Workspace().saveWorkspace(Constants.FALSE)

	#-------------------------------------#
	# set up a "dummy" progress indicator #
	#-------------------------------------#
	progressIndicator = \
			hec.client.RmiProgressImpl(hec.client.ClientApp.app().frame())
	remoteWrapper = hec.server.RemoteWrapper()
	progressIndicator.exportServer()
	remoteWrapper.setExported(Constants.TRUE)
	remoteWrapper.setRemote(progressIndicator)
	progressIndicator.addRemoteWrapper(remoteWrapper)
	progressIndicator.setRunObserver(module)

	#---------------------------------------------------------#
	# tell the forecast to use the "dummy" progress indicator #
	#---------------------------------------------------------#
	forecast.setUserInterface(remoteWrapper)
	forecast.setComputeDisplay(progressIndicator.getDisplay())

	forecast.setComputeAll(Constants.FALSE)
	forecast.setRunsToExecute(alternativesToRun)

#=========================================================================
def getAlternativeNamesAndKeys(forecast) :
	'''
	Associates alternative keys with alternative names
	'''
	dict = {"name" : {}, "key" : {}}
	alternatives = forecast.getAlternatives()
	forecastRuns = forecast.getForecastRuns()
	if not len(alternatives) == len(forecastRuns) :
		# don't really know what to to here
		return None
	for i in range(len(alternatives)) :
		dict["name"][alternatives[i]] = `forecastRuns[i]`
		dict["key"][`forecastRuns[i]`] = alternatives[i]
		dict["key"][alternatives[i]] = alternatives[i]
	return dict
	
#=========================================================================
def makeFPart(modelPosition, alternativeKey, trialNumber = 0) :
	'''
	Builds the F-part of the pathname in the forecast DSS file for the
	given model, alternative and trial.
	
	The model can be specified by name (standard names only) or by its
	ordinal position in the modeling sequence .
	'''
	modelNames = {"MFP" : 1, "HMS" : 2, "RESSIM" : 3, "RAS" : 4, "FIA" : 5}
	if type(modelPosition) == type("") :
		modelPos = modelNames[modelPosition.upper()]
	else :
		modelPos = modelPosition
		
	return `trialNumber`.join(alternativeKey)[:modelPos * 2]

#=========================================================================
def vectorToList(vector) :
	'''
	Utility to make handling returned vectors easier in Jython	
	'''
	if vector == None : return None
	list = []
	for i in range(vector.size()) : list.append(vector.get(i))
	return list

#=========================================================================

