'''
This module allows Jython scripts to access the CWMS database through the 
(D)ata(B)ase (I)nterface.

It is intended to provide essentially the same API as is used to access data
stored in DSS files, although some differences are necessary.

The API is as follows :
	
	==[Functions]==============================================================================

	getVersion()
		Returns the version of this module in the form "X.XX ddMMMyyyy"
		
	openDbi([url [, startTime, endTime]])
		Returns a DbiAccess object.  The object has the default store rule set to 
		"Replace All" , the units system set to "English" and the timeout set to 15 
		seconds.  If the url is None, or not supplied, the object may accesss the DBI 
		connected to the current CAVI session. If the url is supplied, it must be the
		exact string listed in the "Url" column of the "DbiHandler" row on the 
		"Servers" tab of the ServerAdmin dialog 
		(e.g. "rmi://155.83.200.65:64102/DbiHandler"). If startTime and endTime are
		supplied the object will have the spcecified default time window, which can
		be changed by subsequent calls to set setTimeWindow method.  Individual reads
		can override the default time window by specifiying a time window in the
		method call to read or readDataSetTx.
		
	readTSCatalog(fileName [, maxAge])
		Reads a catalog save file and returns the corrsponding TSCatalog object.
		If maxAge is specified, it specifies the valid lifetime of the catalog
		save file in hours.  If it is not specified, the valid lifetime of the 
		catalog save file has no end.
		
	readRatingCatalog(fileName [, maxAge])
		Reads a catalog save file and returns the corrsponding RatingCatalog object.
		If maxAge is specified, it specifies the valid lifetime of the catalog
		save file in hours.  If it is not specified, the valid lifetime of the 
		catalog save file has no end.
		
	dtx2tsm(dtx)
		Returns a TimeSeriesMath object containing the same data as the supplied
		DataSetTx object.  Used internally by DbiAccess.read.
		
	tsm2dtx(tsm)
		Returns a DataSetTx object containing the same data as the supplied
		TimeSeriesMath object.  Used internally by DbiAccess.write.
		
	srs2srm(srs)
		Returns a StreamRatingMath object containing the same data as the supplied
		StreamRatingSeries object.  Used internally by DbiAccess.read.
		
	srm2srs(srm)
		Returns a StreamRatingSeries object containing the same data as the supplied
		StreamRatingMath object.  Used internally by DbiAccess.write.
		
	==[DbiAccess object methods]===============================================================
	
	close()
		Closes a DbiAccess object and releases all its resources.
		
	setTimeWindow(startTime, endTime)
		Sets the default time window for the DbiAccess object.  The default
		time window will be used for all subsequent time-series reads that
		do not in themselves specify a time window. This setting does not
		affect the reading of rating objects.  If a rating is read without
		specifying the time window, the entire time-series of ratings is 
		returned.  If a time window is specified for a rating, only the 
		ratings necessary to rate the specified time window are retrieved.
		
	setStoreRule(storeRule)
		Sets the default store rule for the DbiAccess object.  The 
		default store rule will be used for all subsequent writes
		that do not in themselves specify a store rule.  Valid
		store rules are any of the following strings or constants :
      	
			"Delete Insert"                hec.lang.Const.DELETE_INSERT
			"Replace All"                  hec.lang.Const.REPLACE_ALL
			"Do Not Replace"               hec.lang.Const.DO_NOT_REPLACE
			"Replace Missing Values Only"  hec.lang.Const.REPLACE_MISSING_VALUES_ONLY
			"Replace With Non Missing"     hec.lang.Const.REPLACE_WITH_NON_MISSING
    
	setTimout(seconds)
		Sets the timeout in seconds for the DbiAccess object.  The
		timeout is used in communicating with the CWMS DBI.  Transactions
		are passed to the DBI for processing, and the DBI notifies the 
		originator of the the transaction after it completes.  The getTSCatalog()
		method is immune to the timeout setting, but the read() and write()
		methods respect it.
		
	setUnitsSystem(system)
    		Sets the units system for the DbiAccess object.  The units system
		is used for all subseqent reads.  All data is stored in the CWMS
		database in SI (metric) units.  This call is used to inform the 
		DbiAccess object of which units system for returning data from 
		reads.  Valid units systems are "English" and "SI".

	isConnected()
		Returns whether the object is successfully connected to a Dbi

	canTransact()
		Returns whether the object has sufficient permission (via server ACL)
		to process transactions.  If this value is FALSE, no reads, writes, or
		catalog retrievals can be performed, but the office ID can be retrieved
		via getOfficeID().
		
        getOfficeID()
		Returns the office ID associated with this DBI
		
	getTSCatalog([fileName [,maxAge]])
		Causes the DbiAccessObject to retrieve a catalog of all time series
		identifiers in the database and return the catalog as a list.  If fileName
		is specified, then it is the name of a catalog save file.  If specified,
		the save file is read instead of querying the database, but only if it
		exists, if it is a valid catalog save file, and if it is not out of date.
		If fileName is specified, but a valid, recent save file cannot be read 
		from that file then the database is queried and the resulting catalog is
		written to the save file.  The default valid lifetime of a catalog save
		file is 168 hours (1 week).  If maxAge is specified, it overrides the
		default valid lifetime in hours.  The catalog is returned as a TSCatalog
		object.
	
	getRatingCatalog([fileName [,maxAge]])
		Causes the DbiAccessObject to retrieve a catalog of all rating
		identifiers in the database and return the catalog as a list.  If fileName
		is specified, then it is the name of a catalog save file.  If specified,
		the save file is read instead of querying the database, but only if it
		exists, if it is a valid catalog save file, and if it is not out of date.
		If fileName is specified, but a valid, recent save file cannot be read 
		from that file then the database is queried and the resulting catalog is
		written to the save file.  The default valid lifetime of a catalog save
		file is 168 hours (1 week).  If maxAge is specified, it overrides the
		default valid lifetime in hours.  The catalog is returned as a RatingCatalog
		object.
	
	read(objectID [,startTime, endTime])
		calls readTimeSeriesMath() or readRatingMath(), depending on the format
		of objectID.
	
    	readTimeSeriesMath(timeSeriesID [, startTime, endTime])
    		Causes the DbiAccess object to read time series data from the database 
		and return it in a TimeSeriesMath object.  The timeSeriesID is in CWMS 
		notation (Loc-SubLoc.Param-SubParam.ParamType.Interval.Duration.Version).
		The startTime and endTime, if specified, override the default
		time window setting, but do not replace the default time window
		for subsequent reads.
		
    	readDataSetTx(timeSeriesID [, startTime, endTime])
    		Causes the DbiAccess object to read time series data from the database 
		and return it in a DataSetTx object.  The timeSeriesID is in CWMS 
		notation (Loc[-SubLoc].Param[-SubParam].ParamType.Interval.Duration.Version).
		The startTime and endTime, if specified, override the default
		time window setting, but do not replace the default time window
		for subsequent reads.
		
    	readRatingMath(ratingID [, startTime, endTime])
    		Causes the DbiAccess object to read rating data from the database 
		and return it in a RatingMath object.  The ratingID is in CWMS notation 
		(Loc[-SubLoc].Meas1Param[-subM1][,Meas2Param[-subM2]]:RatedParam[-subR].ratingType.Version).
		The startTime and endTime, if supplied, specify the time window of data to
		be rated, causing the retrieval of only the individual rating records 
		necessary to rate that time window.  If no time window is supplied, the
		entire time series of rating records is retrieved.
		
    	readRatingTimeSeries(ratingID [, startTime, endTime])
    		Causes the DbiAccess object to read rating data from the database 
		and return it in a RatingTimeSeries object.  The ratingID is in CWMS notation 
		(Loc[-SubLoc].Meas1Param[-subM1][,Meas2Param[-subM2]]:RatedParam[-subR].ratingType.Version).
		The startTime and endTime, if supplied, specify the time window of data to
		be rated, causing the retrieval of only the individual rating records 
		necessary to rate that time window.  If no time window is supplied, the
		entire time series of rating records is retrieved.
		
	write(object [, storeRule])
		calls writeTimeSeriesMath(), writeDataSetTx(), writeRatingMath() or 
		writeRatingTimeSeries(), depending on the type of the object supplied.
	
	writeTimeSeriesMath(tsMath [, storeRule])
		Causes the DbiAccess object to write the specified TimeSeriesMath object to 
		the database.  The storeRule, if specifed, overrides the default store rule 
		setting, but does not replace the default store rule for subsequent writes.
			
	writeDataSetTx(dsTx [, storeRule])
		Causes the DbiAccess object to write the specified DataSetTx object to the 
		database.  The storeRule, if specifed, overrides the default store rule 
		setting, but does not replace the default store rule for subsequent writes.
			
	writeRatingMath(rMath [, storeRule])
		Causes the DbiAccess object to write the specified RatingMath object to 
		the database.  The storeRule, if specifed, overrides the default store rule 
		setting, but does not replace the default store rule for subsequent writes.
			
	writeRatingMath(rts [, storeRule])
		Causes the DbiAccess object to write the specified RatingTimeSeries object to 
		the database.  The storeRule, if specifed, overrides the default store rule 
		setting, but does not replace the default store rule for subsequent writes.
			
	==[TSCatalog object methods]===============================================================
	
	save(fileName)
		Saves a copy of the catalog to the specified file.
		
	getTsIds()
		Returns a tuple of all the time series identifiers in the catalog.
		
	getLocations()
		Returns a tuple of all the unique locations in the catalog.

	getSubLocations()
		Returns a tuple of all the unique sub-locations in the catalog.

	getParameters()
		Returns a tuple of all the unique parameters in the catalog.

	getSubParameters()
		Returns a tuple of all the unique sub-parameters in the catalog.
		
	getParameterTypes()
		Returns a tuple of all the unique parameter types in the catalog.

	getIntervals()
		Returns a tuple of all the unique intervals in the catalog.

	getDurations()
		Returns a tuple of all the unique durations in the catalog.

	getVersions()
		Returns a tuple of all the unique versions in the catalog.

	getMatchingTsIds(pattern [, caseSensitive])
		Returns a tuple of all the time series identifiers in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.
		
	getMatchingLocations(pattern [, caseSensitive])
		Returns a tuple of all the unique locations in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingSubLocations(pattern [, caseSensitive])
		Returns a tuple of all the unique sub-locations in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingParameters(pattern [, caseSensitive])
		Returns a tuple of all the unique parameters in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingSubParameters(pattern [, caseSensitive])
		Returns a tuple of all the unique sub-parameters in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.
		
	getMatchingParameterTypes(pattern [, caseSensitive])
		Returns a tuple of all the unique parameter types in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingIntervals(pattern [, caseSensitive])
		Returns a tuple of all the unique intervals in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingDurations(pattern [, caseSensitive])
		Returns a tuple of all the unique durations in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingVersions(pattern [, caseSensitive])
		Returns a tuple of all the unique versions in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	==[RatingCatalog object methods]===========================================================
	
	save(fileName)
		Saves a copy of the catalog to the specified file.
		
	getRatingIds()
		Returns a tuple of all the rating identifiers in the catalog.
		
	getLocations()
		Returns a tuple of all the unique locations in the catalog.

	getSubLocations()
		Returns a tuple of all the unique sub-locations in the catalog.

	getMeasured1Parameters()
		Returns a tuple of all the unique measured_1 parameters in the catalog.

	getMeasured1SubParameters()
		Returns a tuple of all the unique measured_1 sub-parameters in the catalog.
		
	getMeasured2Parameters()
		Returns a tuple of all the unique measured_2 parameters in the catalog.

	getMeasured2SubParameters()
		Returns a tuple of all the unique measured_2 sub-parameters in the catalog.
		
	getRatedParameters()
		Returns a tuple of all the unique rated parameters in the catalog.

	getRatedSubParameters()
		Returns a tuple of all the unique rated sub-parameters in the catalog.
		
	getRatingTypes()
		Returns a tuple of all the unique rating types in the catalog.

	getVersions()
		Returns a tuple of all the unique versions in the catalog.

	getMatchingRatingIds(pattern [, caseSensitive])
		Returns a tuple of all the time series identifiers in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.
		
	getMatchingLocations(pattern [, caseSensitive])
		Returns a tuple of all the unique locations in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingSubLocations(pattern [, caseSensitive])
		Returns a tuple of all the unique sub-locations in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingMeasured1Parameters(pattern [, caseSensitive])
		Returns a tuple of all the unique measured_1 parameters in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingMeasured1SubParameters(pattern [, caseSensitive])
		Returns a tuple of all the unique measured_1 sub-parameters in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.
		
	getMatchingMeasured2Parameters(pattern [, caseSensitive])
		Returns a tuple of all the unique measured_2 parameters in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingMeasured2SubParameters(pattern [, caseSensitive])
		Returns a tuple of all the unique measured_2 sub-parameters in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.
		
	getMatchingRatedParameters(pattern [, caseSensitive])
		Returns a tuple of all the unique rated parameters in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingRatedSubParameters(pattern [, caseSensitive])
		Returns a tuple of all the unique rated sub-parameters in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.
		
	getMatchingRatingTypes(pattern [, caseSensitive])
		Returns a tuple of all the unique rating types in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.

	getMatchingVersions(pattern [, caseSensitive])
		Returns a tuple of all the unique versions in the catalog that match the 
		specified pattern.  Pattern matching is performed using file name matching
		conventions, where the '*'character matches zero or more occurences of any 
		character and the '?' character matches zero or one occurence of any character.
		By default, the matching is performed in a case-insensitive manner, which can
		be altered by supplying a non-zero value for caseSensitive.


Mike Perryman
USACE HEC
Aug 2004

History

	1.11	10Sep2004	Added setName() DbiAccess method.  Modified dtx2tsm() function
				to convert the interval field in the TimeSeriesDescription
				from seconds to minutes to account for a (now entrenched) bug
				in hec.data.Tx.DataSetTx.getTimeSeriesContainer()
	
	1.10	26Aug2004	Added getVersion() function.  Also added getOfficeID(), 
				isConnected(), and canTransact() DbiAccess object methods.  
				Modified openDbi() function to recognize connections
				to DBIs for which the client does not have transaction 
				permissions.  Fixed bug in DbiAccess.setUnitsSystem().
				
	1.00	16Jun2004	Original version
'''

import hec, wcds, rma, java, time, thread, fnmatch, sys

TRUE, FALSE = 1, 0
##############################################################################

class reentrantLock :
	
	'''
	Mimic Python's Threading.RLock class in Jython.
	'''
	
	def __init__(self) :
		'''Construct a reentrantLock object.'''
		self._lock   = thread.allocate_lock()
		self._count  = 0
		self._thread = None
		
	def locked(self) :
		'''Return whether this object is locked.'''
		return self._lock.locked()
		
	def acquire(self, waitflag=None) :
		'''Acquire (lock) this object, with optional timeout.'''
		thisThread = thread.get_ident()
		if self._lock.locked() :
			if self._thread == thisThread :
				self._count += 1
				return None
			else :
				if waitflag == None :
					self._lock.acquire()
				else :
					rc = self._lock.acquire(waitflag)
				if self._lock.locked() and self._count == 0 and self._thread == None :
					self._thread = thisThread
					self._count = 1
					if waitflag == None : 
						return None
					else :
						return rc
		else :
			if waitflag == None :
				self._lock.acquire()
			else :
				self._lock.acquire(waitflag)
			self._thread = thisThread
			self._count = 1
			
	def release(self) :
		'''Release (unlock) this object.'''
		thisThread = thread.get_ident()
		if self._lock.locked() :
			if self._thread == thisThread :
				self._count -= 1
				if self._count == 0:
					self._thread = None
					self._lock.release()
		
##############################################################################
class TSCatalog :
	
	'''
	Class that holds a time series catalog list and allows getting lists
	of the uniques pieces and matching partial strings.
	'''
	
	def __init__(self, list, parts=None) :
		'''
		Constuct a TSCatalog object.
		'''
		self._list = list
		self._locations = []
		self._subLocations = []
		self._parameters = []
		self._subParameters = []
		self._parameterTypes = []
		self._intervals = []
		self._durations = []
		self._versions = []
		if parts :
			for i in range(len(parts)) :
				if parts[i] == None : parts[i] = []
			self._locations      = parts[0]
			self._subLocations   = parts[1]
			self._parameters     = parts[2]
			self._subParameters  = parts[3]
			self._parameterTypes = parts[4]
			self._intervals      = parts[5]
			self._durations      = parts[6]
			self._versions       = parts[7]
		else :
			for tsid in list :
				tsid = tsid.strip()
				loc, param, paramType, intvl, dur, ver = tsid.split(".")
				try    : loc, subLoc = loc.split("-", 1)
				except : subLoc = None
				try    : param, subParam = param.split("-", 1)
				except : subParam = None
				if not loc in self._locations :
					self._locations.append(loc)
				if subLoc and subLoc not in self._subLocations :
					self._subLocations.append(subLoc)
				if not param in self._parameters :
					self._parameters.append(param)
				if subParam and subParam not in self._subParameters :
					self._subParameters.append(subParam)
				if not paramType in self._parameterTypes :
					self._parameterTypes.append(paramType)
				if not intvl in self._intervals :
					self._intervals.append(intvl)
				if not dur in self._durations :
					self._durations.append(dur)
				if not ver in self._versions :
					self._versions.append(ver)
					
	def save(self, fileName) :
		'''
		Save catalog data to disk.
		'''
		rc = 0
		try :
			file = open(fileName, "w")
			file.write("%s\n" % TSCatalogMagicString);
			file.write("%f\n" % time.time())
			for tsId in self._list : 
				file.write("%s\n" % tsId)
			file.write("%s\n" % TSCatalogMagicString);
			for item in self._locations : 
				file.write("%s\n" % item)
			file.write("%s\n" % TSCatalogMagicString);
			for item in self._subLocations :
				file.write("%s\n" % item)
			file.write("%s\n" % TSCatalogMagicString);
			for item in self._parameters : 
				file.write("%s\n" % item)
			file.write("%s\n" % TSCatalogMagicString);
			for item in self._subParameters : 
				file.write("%s\n" % item)
			file.write("%s\n" % TSCatalogMagicString);
			for item in self._parameterTypes : 
				file.write("%s\n" % item)
			file.write("%s\n" % TSCatalogMagicString);
			for item in self._intervals : 	
				file.write("%s\n" % item)
			file.write("%s\n" % TSCatalogMagicString);
			for item in self._durations : 
				file.write("%s\n" % item)
			file.write("%s\n" % TSCatalogMagicString);
			for item in self._versions : 
				file.write("%s\n" % item)
			file.close()
			rc = 1
		except :
			pass
		return rc
		
	def getMatchingLocations(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog locations matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for location in self._locations :
				if fnmatch.fnmatchcase(location, pattern) :
					matches.append(location)
		else :
			patternlwr = pattern.lower()
			for location in self._locations :
				if fnmatch.fnmatch(location.lower(), patternlwr) :
					matches.append(location)
		return tuple(matches)
		
	def getMatchingSubLocations(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog sub-locations matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for subLocation in self._subLocations :
				if fnmatch.fnmatchcase(subLocation, pattern) :
					matches.append(subLocation)
		else :
			patternlwr = pattern.lower()
			for subLocation in self._subLocations :
				if fnmatch.fnmatch(subLocation.lower(), patternlwr) :
					matches.append(subLocation)
		return tuple(matches)
		
	def getMatchingParameters(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog parameters matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for parameter in self._parameters :
				if fnmatch.fnmatchcase(parameter, pattern) :
					matches.append(parameter)
		else :
			patternlwr = pattern.lower()
			for parameter in self._parameters :
				if fnmatch.fnmatch(parameter.lower(), patternlwr) :
					matches.append(parameter)
		return tuple(matches)
		
	def getMatchingSubParameters(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog sub-parameters matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for subParameter in self._subParameters :
				if fnmatch.fnmatchcase(subParameter, pattern) :
					matches.append(subParameter)
		else :
			patternlwr = pattern.lower()
			for subParameter in self._subParameters :
				if fnmatch.fnmatch(subParameter.lower(), patternlwr) :
					matches.append(subParameter)
		return tuple(matches)
		
	def getMatchingParameterTypes(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog parameter types matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for parameterType in self._parameterTypes :
				if fnmatch.fnmatchcase(parameterType, pattern) :
					matches.append(parameterType)
		else :
			patternlwr = pattern.lower()
			for parameterType in self._parameterTypes :
				if fnmatch.fnmatch(parameterType.lower(), patternlwr) :
					matches.append(parameterType)
		return tuple(matches)
		
	def getMatchingIntervals(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog intervals matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for interval in self._intervals :
				if fnmatch.fnmatchcase(interval, pattern) :
					matches.append(interval)
		else :
			patternlwr = pattern.lower()
			for interval in self._intervals :
				if fnmatch.fnmatch(interval.lower(), patternlwr) :
					matches.append(interval)
		return tuple(matches)
		
	def getMatchingDurations(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog durations matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for duration in self._durations :
				if fnmatch.fnmatchcase(duration, pattern) :
					matches.append(duration)
		else :
			patternlwr = pattern.lower()
			for duration in self._durations :
				if fnmatch.fnmatch(duration.lower(), patternlwr) :
					matches.append(duration)
		return tuple(matches)
		
	def getMatchingVersions(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog versions matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for version in self._versions :
				if fnmatch.fnmatchcase(version, pattern) :
					matches.append(version)
		else :
			patternlwr = pattern.lower()
			for version in self._versions :
				if fnmatch.fnmatch(version.lower(), patternlwr) :
					matches.append(version)
		return tuple(matches)
		
	def getMatchingTsIds(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog time-series ids matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for tsid in self._list :
				if fnmatch.fnmatchcase(tsid, pattern) :
					matches.append(tsid)
		else :
			patternlwr = pattern.lower()
			for tsid in self._list :
				if fnmatch.fnmatch(tsid.lower(), patternlwr) :
					matches.append(tsid)
		return tuple(matches)
		
	def getLocations(self) :
		'''
		Return a tuple of all catalog locations.
		'''
		return tuple(self._locations)
	
	def getSubLocations(self) :
		'''
		Return a tuple of all catalog sub-locations.
		'''
		return tuple(self._subLocations)
	
	def getParameters(self) :
		'''
		Return a tuple of all catalog parameters.
		'''
		return tuple(self._parameters)
	
	def getSubParameters(self) :
		'''
		Return a tuple of all catalog sub-parameters.
		'''
		return tuple(self._parameters)
	
	def getParameterTypes(self) :
		'''
		Return a tuple of all catalog parameter types.
		'''
		return tuple(self._parameterTypes)
	
	def getIntervals(self) :
		'''
		Return a tuple of all catalog intervals.
		'''
		return tuple(self._intervals)
	
	def getDurations(self) :
		'''
		Return a tuple of all catalog durations.
		'''
		return tuple(self._durations)
	
	def getVersions(self) :
		'''
		Return a tuple of all catalog versions.
		'''
		return tuple(self._versions)
	
	def getTsIds(self) :
		'''
		Return a tuple of all catalog time-series identifiers.
		'''
		return tuple(self._list)
	
##############################################################################

class RatingCatalog :
	
	'''
	Class that holds a rating catalog list and allows getting lists
	of the uniques pieces and matching partial strings.
	'''
	
	def __init__(self, list, parts=None) :
		'''
		Constuct a RatingCatalog object.
		'''
		self._list = list
		self._locations = []
		self._subLocations = []
		self._meas1Params = []
		self._meas1SubParams = []
		self._meas2Params = []
		self._meas2SubParams = []
		self._ratedParams = []
		self._ratedSubParams = []
		self._ratingTypes = []
		self._versions = []
		if parts :
			for i in range(len(parts)) :
				if parts[i] == None : parts[i] = []
			self._locations      = parts[0]
			self._subLocations   = parts[1]
			self._meas1Params    = parts[2]
			self._meas1SubParams = parts[3]
			self._meas2Params    = parts[4]
			self._meas2SubParams = parts[5]
			self._ratedParams    = parts[6]
			self._ratedSubParams = parts[7]
			self._ratingTypes    = parts[8]
			self._versions       = parts[9]
			for i in range(len(self._ratingTypes)) :
				self._ratingTypes[i] = "Rating%s" % self._ratingTypes[i].title()
		else :
			for tsid in list :
				tsid = tsid.strip()
				loc, params, rType, ver = tsid.split(".")
				meas1Param, ratedParam = params.split(":")
				try    : meas1Param, meas2Param = meas1Param.split(",")
				except : meas2Param = None
				try    : loc, subLoc = loc.split("-", 1)
				except : subLoc = None
				try    : meas1Param, meas1SubParam = meas1Param.split("-", 1)
				except : meas1SubParam = None
				try    : meas2Param, meas2SubParam = meas2Param.split("-", 1)
				except : meas2SubParam = None
				try    : ratedParam, ratedSubParam = ratedParam.split("-", 1)
				except : ratedSubParam = None
				
				if not loc in self._locations : self._locations.append(loc)
				if subLoc and subLoc not in self._subLocations : self._subLocations.append(subLoc)
				if not meas1Param in self._meas1Params : self._meas1Params.append(meas1Param)
				if meas1SubParam and meas1SubParam not in self._meas1SubParams : self._meas1SubParams.append(meas1SubParam)
				if meas2Param :
					if not meas2Param in self._meas2Params : self._meas2Params.append(meas2Param)
					if meas2SubParam and meas2SubParam not in self._meas2SubParams : self._meas2SubParams.append(meas2SubParam)
				if not ratedParam in self._ratedParams : self._ratedParams.append(ratedParam)
				if ratedSubParam and ratedSubParam not in self._ratedSubParams : self._ratedSubParams.append(ratedSubParam)
				if not ratingType in self._ratingTypes : self._ratingTypes.append(ratingType)
				if not ver in self._versions : self._versions.append(version)
					
	def save(self, fileName) :
		'''
		Save catalog data to disk.
		'''
		rc = 0
		try :
			file = open(fileName, "w")
			file.write("%s\n" % RatingCatalogMagicString);
			file.write("%f\n" % time.time())
			for tsId in self._list : file.write("%s\n" % tsId)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._locations : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._subLocations : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._meas1Params : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._meas1SubParams : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._meas2Params : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._meas2SubParams : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._ratedParams : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._ratedSubParams : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._ratingTypes : file.write("%s\n" % item)
			file.write("%s\n" % RatingCatalogMagicString);
			for item in self._versions : file.write("%s\n" % item)
			file.close()
			rc = 1
		except :
			pass
		return rc
		
	def getMatchingLocations(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog locations matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for location in self._locations :
				if fnmatch.fnmatchcase(location, pattern) :
					matches.append(location)
		else :
			patternlwr = pattern.lower()
			for location in self._locations :
				if fnmatch.fnmatch(location.lower(), patternlwr) :
					matches.append(location)
		return tuple(matches)
		
	def getMatchingSubLocations(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog sub-locations matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for subLocation in self._subLocations :
				if fnmatch.fnmatchcase(subLocation, pattern) :
					matches.append(subLocation)
		else :
			patternlwr = pattern.lower()
			for subLocation in self._subLocations :
				if fnmatch.fnmatch(subLocation.lower(), patternlwr) :
					matches.append(subLocation)
		return tuple(matches)
		
	def getMatchingMeasured1Parameters(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog measured_1 parameters matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for parameter in self._meas1Params :
				if fnmatch.fnmatchcase(parameter, pattern) :
					matches.append(parameter)
		else :
			patternlwr = pattern.lower()
			for parameter in self._meas1Params :
				if fnmatch.fnmatch(parameter.lower(), patternlwr) :
					matches.append(parameter)
		return tuple(matches)
		
	def getMatchingMeasured1SubParameters(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog measured_1 sub-parameters matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for subParameter in self._meas1SubParams :
				if fnmatch.fnmatchcase(subParameter, pattern) :
					matches.append(subParameter)
		else :
			patternlwr = pattern.lower()
			for subParameter in self._meas1SubParams :
				if fnmatch.fnmatch(subParameter.lower(), patternlwr) :
					matches.append(subParameter)
		return tuple(matches)
		
	def getMatchingMeasured2Parameters(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog measured_2 parameters matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for parameter in self._meas2Params :
				if fnmatch.fnmatchcase(parameter, pattern) :
					matches.append(parameter)
		else :
			patternlwr = pattern.lower()
			for parameter in self._meas2Params :
				if fnmatch.fnmatch(parameter.lower(), patternlwr) :
					matches.append(parameter)
		return tuple(matches)
		
	def getMatchingMeasured2SubParameters(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog measured_2 sub-parameters matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for subParameter in self._meas2SubParams :
				if fnmatch.fnmatchcase(subParameter, pattern) :
					matches.append(subParameter)
		else :
			patternlwr = pattern.lower()
			for subParameter in self._meas2SubParams :
				if fnmatch.fnmatch(subParameter.lower(), patternlwr) :
					matches.append(subParameter)
		return tuple(matches)
		
	def getMatchingRatedParameters(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog rated parameters matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for parameter in self._ratedParams :
				if fnmatch.fnmatchcase(parameter, pattern) :
					matches.append(parameter)
		else :
			patternlwr = pattern.lower()
			for parameter in self._ratedParams :
				if fnmatch.fnmatch(parameter.lower(), patternlwr) :
					matches.append(parameter)
		return tuple(matches)
		
	def getMatchingRatedSubParameters(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog rated sub-parameters matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for subParameter in self._ratedSubParams :
				if fnmatch.fnmatchcase(subParameter, pattern) :
					matches.append(subParameter)
		else :
			patternlwr = pattern.lower()
			for subParameter in self._ratedSubParams :
				if fnmatch.fnmatch(subParameter.lower(), patternlwr) :
					matches.append(subParameter)
		return tuple(matches)
		
	def getMatchingRatingTypes(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog rating types matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for rType in self._ratingTypes :
				if fnmatch.fnmatchcase(rType, pattern) :
					matches.append(rType)
		else :
			patternlwr = pattern.lower()
			for rType in self._ratingTypes :
				if fnmatch.fnmatch(rType.lower(), patternlwr) :
					matches.append(rType)
		return tuple(matches)
		
	def getMatchingVersions(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog versions matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for version in self._versions :
				if fnmatch.fnmatchcase(version, pattern) :
					matches.append(version)
		else :
			patternlwr = pattern.lower()
			for version in self._versions :
				if fnmatch.fnmatch(version.lower(), patternlwr) :
					matches.append(version)
		return tuple(matches)
		
	def getMatchingRatingIds(self, pattern, caseSensitive=0) :
		'''
		Return a tuple of catalog time-series ids matching a specified pattern.
		'''
		matches = []
		if caseSensitive :
			for tsid in self._list :
				if fnmatch.fnmatchcase(tsid, pattern) :
					matches.append(tsid)
		else :
			patternlwr = pattern.lower()
			for tsid in self._list :
				if fnmatch.fnmatch(tsid.lower(), patternlwr) :
					matches.append(tsid)
		return tuple(matches)
		
	def getLocations(self) :
		'''
		Return a tuple of all catalog locations.
		'''
		return tuple(self._locations)
	
	def getSubLocations(self) :
		'''
		Return a tuple of all catalog sub-locations.
		'''
		return tuple(self._subLocations)
	
	def getMeasured1Parameters(self) :
		'''
		Return a tuple of all catalog measured_1 parameters.
		'''
		return tuple(self._meas1Params)
	
	def getMeasured1SubParameters(self) :
		'''
		Return a tuple of all catalog measured_1 sub-parameters.
		'''
		return tuple(self._meas1SubParams)
	
	def getMeasured2Parameters(self) :
		'''
		Return a tuple of all catalog measured_2 parameters.
		'''
		return tuple(self._meas2Params)
	
	def getMeasured2SubParameters(self) :
		'''
		Return a tuple of all catalog measured_2 sub-parameters.
		'''
		return tuple(self._meas2SubParams)
	
	def getRatedParameters(self) :
		'''
		Return a tuple of all catalog rated parameters.
		'''
		return tuple(self._ratedParams)
	
	def getRatedSubParameters(self) :
		'''
		Return a tuple of all catalog rated sub-parameters.
		'''
		return tuple(self._ratedSubParams)
	
	def getVersions(self) :
		'''
		Return a tuple of all catalog versions.
		'''
		return tuple(self._versions)
	
	def getRatingIds(self) :
		'''
		Return a tuple of all catalog rating identifiers.
		'''
		return tuple(self._list)
	
##############################################################################

class DbiAccess(java.util.Observer) :
	
	'''
	Class that paralells the DssFile object for access to the
	database via the DBI, plus extra functionality.
	'''
	
	def __init__(self, url=None, startTime=None, endTime=None) :
		'''
		Construct a DbiAccess object.
		'''
		self._clientLock = reentrantLock()
		self._clientLock.acquire()
		self._initValues()
		self._callbackLock = reentrantLock()
		self._catalogLock= reentrantLock()

		if not url :
			self._dbiWrapper = hec.client.ClientApp.app().getDbi()
		else :
			self._dbiUrl = url
			self._dbiWrapper = wcds.dbi.RmiDbiWrapper(url, rma.util.Priorities.NORM_PRIORITY+1)

		self._canTransact = self._dbiWrapper.getDbiStatus().find("rejected") == -1
		if not self._canTransact:
			sys.stderr.write("Cannot transact with Dbi.\n")
		self.setTimeWindow(startTime, endTime)
		self._clientLock.release()
		
	def _initValues(self) :
		'''
		Initialize the member variable values.
		'''		
		self._dbiUrl          = None
		self._startTimeStr    = None
		self._startTime       = None
		self._endTimeStr      = None
		self._endTime         = None
		self._transactionTime = None
		self._dbiWrapper      = None
		self._canTransact     = FALSE
		self._dataApp         = None
		self._unitsSystem     = "English"
		self._timeout         = 60 # seconds
		self._startAction     = None
		self._endAction       = None
		self._storeRule       = hec.lang.Const.REPLACE_ALL
		self._timezone        = java.util.TimeZone.getTimeZone("UTC")
		self._returnValue     = None
		self._name            = "DBI Jython Module"

	def _checkStatus(self) :
		'''
		Make sure we can talk to the Dbi
		'''
		if not self._dbiWrapper :
			self._clientLock.release()
			raise SystemError, "The object is not connected to a Dbi."
		if not self._canTransact :
			self._clientLock.release()
			raise SystemError, "Cannot transact with Dbi."

	def isConnected(self) :
		'''
		Returns whether we are connected.
		'''
		self._clientLock.acquire()
		is_connected = self._dbiWrapper and self._dbiWrapper.isConnected()
		self._clientLock.release()
		return is_connected

	def canTransact(self) :
		'''
		Returns whether we have permission to transact with Dbi
		'''
		self._clientLock.acquire()
		canTransact = self._canTransact
		self._clientLock.release()
		return canTransact

	def setName(self, name) :
		'''Set the name used to identify this object in the DBI log
		'''
		self._name = name
		
	def setTimeWindow(self, startTimeStr, endTimeStr) :
		'''
		Set the default time window for future dbi accesses.
		'''
		self._clientLock.acquire()
		if startTimeStr and endTimeStr :
			startTime = rma.util.RMAIO.parseDate(startTimeStr, self._timezone)
			endTime = rma.util.RMAIO.parseDate(endTimeStr, self._timezone)
			self._startTimeStr = startTimeStr
			self._startTime = startTime
			self._endTimeStr = endTimeStr
			self._endTime = endTime
		elif startTimeStr or endTimeStr :
			self._clientLock.release()
			raise ValueError, "Both start and end times must be specified"
		self._clientLock.release()
	
	def setUnitsSystem(self, system) :
		'''
		Set the working units system to SI or English, used for reads only.
		'''
		self._clientLock.acquire()
		if system.lower() == "english" :
			self._unitsSystem = "English"
		elif system.lower() == "si" :
			self._unitsSystem = "SI"
		else :
			self._clientLock.release()
			raise ValueError, "System must be Enlish or SI."
		self._clientLock.release()
			
	def setTimeout(self, seconds) :
		'''
		Sets the timeout in seconds for waiting on callback from the DBI.
		'''
		if type(seconds) != type(0) :
			raise TypeError, "setTimeout parameter must be an integer."
		self._clientLock.acquire()
		self._timeout = seconds
		self._clientLock.release()

	def setStoreRule(self, storeRule) :
		'''
		Sets the default store rule, used for writes only.
		'''
		self._clientLock.acquire()
		try :
			self._storeRule = getStoreRuleNumber(storeRule)
		except :
			pass
		self._clientLock.release()
			
        def getOfficeID(self) :
		'''
		Return the office ID associated with this DBI
		'''
		self._clientLock.acquire()
		officeId = self._dbiWrapper.getOfficeId()
		self._clientLock.release()
		return officeId

	def getTSCatalog(self, fileName=None, maxAge=168) :
		'''
		Retrieve a catalog of all time-series ids in the database for this office id.
		'''
		#---------------------------------------------------------#
		# read the catalog from the save file if one is specified #
		#---------------------------------------------------------#
		if fileName :
			catalog = readTSCatalog(fileName, maxAge)
			if catalog : 
				sys.stderr.write("Catalog restored from file %s.\n" % fileName)
				return catalog
		#-------------#
		# set up shop #
		#-------------#
		template = hec.data.tx.DescriptionTx.getDescriptionTxCatalogTemplate()
		self._clientLock.acquire()
		self._checkStatus()
		if not self._dataApp :
			self._dataApp = wcds.data.DbiDataApp.buildApp(
				self._dbiWrapper.getOfficeId(), 
				"Jython object %s" % self.__class__, 
				self._dbiWrapper)
		# self._clientLock.acquire()
		#--------------------#
		# query the database #
		#--------------------#
		self._catalogLock.acquire()
		t1 = time.time()
		dbCat = self._dataApp.retrieveCatalog(None, template)
		t2 = time.time()
		self._catalogLock.release()
		available = dbCat.getRows();
		sys.stderr.write(
			"Catalog returned %d records from database in %.3f seconds, processing..." % 
			(available.size(), t2 - t1))
		#-------------------#
		# build the catalog #
		#-------------------#
		list = []
		locations = []
		subLocations = []
		parameters = []
		subParameters = []
		parameterTypes = []
		intervals = []
		durations = []
		versions = []
		tmp = java.util.ArrayList(available.size());
		for i in range(available.size()):
			try :
				row = available.get(i);
				location = row.get(0);
				subLocation = row.get(1);
				parameter = row.get(2);
				subParameter = row.get(3);
				parameterType = row.get(4);
				interval = row.get(5);
				duration = row.get(6);
				version = row.get(7);
			except :
				sys.stderr.write("Error processing entry %d (%s).\n" % (i, row))
				continue
			id = location
			if subLocation : id += '-%s' % subLocation
			id += '.%s' % parameter
			if subParameter : id += '-%s' % subParameter
			id += '.%s.%s.%s.%s' % (parameterType, interval, duration, version)
			list.append(id)
			if not location in locations :
				locations.append(location)
			if subLocation and subLocation not in subLocations :
				subLocations.append(subLocation)
			if not parameter in parameters :
				parameters.append(parameter)
			if subParameter and subParameter not in subParameters :
				subParameters.append(subParameter)
			if not parameterType in parameterTypes :
				parameterTypes.append(parameterType)
			if not interval in intervals :
				intervals.append(interval)
			if not duration in durations :
				durations.append(duration)
			if not version in versions :
				versions.append(version)
		parts = [locations, subLocations, parameters, subParameters, parameterTypes, intervals, durations, versions]
		sys.stderr.write("done\n")
		catalog = TSCatalog(list, parts)
		#------------------------------------------#
		# save the catalog, if a name was supplied #
		#------------------------------------------#
		if fileName : catalog.save(fileName)
		return catalog
		
	def getRatingCatalog(self, fileName=None, maxAge=168) :
		'''
		Retrieve a catalog of all time-series ids in the database for this office id.
		'''
		#---------------------------------------------------------#
		# read the catalog from the save file if one is specified #
		#---------------------------------------------------------#
		if fileName :
			catalog = readRatingCatalog(fileName, maxAge)
			if catalog : 
				sys.stderr.write("Catalog restored from file %s.\n" % fileName)
				return catalog
		#-------------#
		# set up shop #
		#-------------#
		template = hec.data.rating.RatingInformation.getCatalogTemplate()
		self._clientLock.acquire()
		self._checkStatus()
		if not self._dataApp :
			self._dataApp = wcds.data.DbiDataApp.buildApp(
				self._dbiWrapper.getOfficeId(), 
				"Jython object %s" % self.__class__, 
				self._dbiWrapper)
		#self._clientLock.acquire()
		#--------------------#
		# query the database #
		#--------------------#
		self._catalogLock.acquire()
		t1 = time.time()
		dbCat = self._dataApp.retrieveCatalog(None, template)
		t2 = time.time()
		self._catalogLock.release()
		available = dbCat.getRows();
		sys.stderr.write(
			"Catalog returned %d records from database in %.3f seconds, processing..." % 
			(available.size(), t2 - t1))
		#-------------------#
		# build the catalog #
		#-------------------#
		locations = []
		subLocations = []
		meas1Params = []
		subMeas1Params = []
		meas2Params = []
		subMeas2Params = []
		ratedParams = []
		subRatedParams = []
		ratingTypes = []
		versions = []
		descriptors = []
		tmp = java.util.ArrayList(available.size());
		for i in range(available.size()):
			try :
				row = available.get(i)
				code = row.get(0)
				office = row.get(1)
				location = row.get(2)
				subLocation = row.get(3)
				meas1Param = row.get(4)
				subMeas1Param = row.get(5)
				meas2Param = row.get(6)
				subMeas2Param = row.get(7)
				ratedParam = row.get(8)
				subRatedParam = row.get(9)
				ratingType = row.get(10)
				version = row.get(11)
				descriptor = row.get(12)
			except :
				sys.stderr.write("Error processing entry %d (%s).\n" % (i, row))
				continue
			if not location in locations : locations.append(location)
			if not subLocation in subLocations : subLocations.append(subLocation)
			if not meas1Param in meas1Params : meas1Params.append(meas1Param)
			if not subMeas1Param in subMeas1Params : subMeas1Params.append(subMeas1Param)
			if not meas2Param in meas2Params : meas2Params.append(meas2Param)
			if not subMeas2Param in subMeas2Params : subMeas2Params.append(subMeas2Param)
			if not ratedParam in ratedParams : ratedParams.append(ratedParam)
			if not subRatedParam in subRatedParams : subRatedParams.append(subRatedParam)
			if not ratingType in ratingTypes : ratingTypes.append(ratingType)
			if not version in versions : versions.append(version)
			if not descriptor in descriptors : descriptors.append(descriptor)
		parts = [locations, subLocations, meas1Params, subMeas1Params, meas2Params, 
			subMeas2Params, ratedParams, subRatedParams, ratingTypes, versions]
		sys.stderr.write("done\n")
		catalog = RatingCatalog(descriptors, parts)
		#------------------------------------------#
		# save the catalog, if a name was supplied #
		#------------------------------------------#
		if fileName : catalog.save(fileName)
		return catalog
		
        def read(self, id, startTimeStr=None, endTimeStr=None):
		'''
		Read a time-series or a rating object from the database and return it as an HecMath object.
		'''
		fields = id.split(".")
		if len(fields) == 6 :
			return self.readTimeSeriesMath(id, startTimeStr, endTimeStr)
		if len(fields) == 4 :
			return self.readRatingMath(id, startTimeStr, endTimeStr)
		raise ValueError, '"%s" is neither a time-series id nor a rating id.' % id
		 
        def readTimeSeriesMath(self, tsId, startTimeStr=None, endTimeStr=None):
		'''
		Read a time-series from the database and return it in a TimeSeriesMath object.
		'''
		return dtx2tsm(self.readDataSetTx(tsId, startTimeStr, endTimeStr));
		
        def readDataSetTx(self, tsId, startTimeStr=None, endTimeStr=None) :
		'''
		Read a time-series from the database and return it in a DataSetTx object.
		'''
		self._clientLock.acquire()
		self._checkStatus()
		#------------------------#
		# handle the time window #
		#------------------------#
		if startTimeStr and endTimeStr :
			startTime = rma.util.RMAIO.parseDate(startTimeStr, self._timezone)
			endTime = rma.util.RMAIO.parseDate(endTimeStr, self._timezone)
			timeWindow = "%s %s" % (startTimeStr, endTimeStr)
		elif startTimeStr or endTimeStr :
			self._clientLock.release()
			raise ValueError, "Both start and end times must be specified, or neither."
		else :
			if not self._startTime and not self._endTime :
				self._clientLock.release()
				raise ValueError, "Time window is not specified, and has not previously been set."
			startTime = self._startTime
			endTime = self._endTime
			timeWindow = "%s %s" % (self._startTimeStr, self._endTimeStr)
		#-----------------------------------------------#
		# create a the structures and submit to the DBI #
		#-----------------------------------------------#
		desc = hec.data.tx.DescriptionTx(tsId)
		dstxTemplate = hec.data.tx.DataSetTxTemplate(desc, startTime, endTime, hec.data.Units())
		action = wcds.lang.Action(dstxTemplate, 
			desc, 
			wcds.lang.Action.RETRIEVE, 
			"DBI", 
			self._name)
		action.setDescription("Retrieving %s , TW=[%s]" % (tsId, timeWindow))
		self.transactWithDBI(action)
		if not self._data : 
			# sys.stderr.write("No values were returned for action : %s.\n" % action.getDescription())
			self._clientLock.release()
			return None
		if self._unitsSystem == "English" :
			self._data.changeUnitSystem(hec.heclib.util.Unit.ENGLISH_ID)
		self._clientLock.release()
		return self._data
		
        def readRatingMath(self, ratingId, startTimeStr=None, endTimeStr=None) :
		'''
		Read a rating from the database and return it in a StreamRatingMath object.
		'''
		return srs2srm(self.readRatingTimeSeries(ratingId, startTimeStr, endTimeStr))
		
        def readRatingTimeSeries (self, ratingId, startTimeStr=None, endTimeStr=None) :
		'''
		Read a rating from the database and return it in a StreamRatingSeries object.
		'''
		self._clientLock.acquire()
		self._checkStatus()
		#------------------------#
		# handle the time window #
		#------------------------#
		if startTimeStr and endTimeStr :
			startTime = rma.util.RMAIO.parseDate(startTimeStr, self._timezone)
			endTime = rma.util.RMAIO.parseDate(endTimeStr, self._timezone)
			timeWindow = "%s %s" % (startTimeStr, endTimeStr)
		elif startTimeStr or endTimeStr :
			self._clientLock.release()
			raise ValueError, "Both start and end times must be specified, or neither."
		else :
			startTime = hec.lang.Const.UNDEFINED_LONG
			endTime = hec.lang.Const.UNDEFINED_LONG
		#-----------------------------------------------#
		# create a the structures and submit to the DBI #
		#-----------------------------------------------#
		actionSpecifier = hec.data.rating.ActionSpecifier(ratingId, 0, startTime, endTime, None)
		action = actionSpecifier.getAction("DBI", self._name)
		if startTimeStr :
			action.setDescription("Retrieving %s , TW=[%s]" % (ratingId, timeWindow))
		else :
			action.setDescription("Retrieving %s (no time window)" % ratingId)
		self.transactWithDBI(action)
		if not self._data : 
			# sys.stderr.write("No rating was returned for action : %s.\n" % action.getDescription())
			self._clientLock.release()
			return None
		if self._unitsSystem == "English" :
			self._data.convertToEnglishUnits()
		self._clientLock.release()
		return self._data
		
        def write(self, object, storeRule=None) :
		'''
		Write an object to the database.
		'''
		if   issubclass(object.__class__, hec.hecmath.TimeSeriesMath)       : self.writeTimeSeries(object, storeRule)
		elif issubclass(object.__class__, hec.data.rating.RatingTimeSeries) : self.writeRating(object, storeRule)
		elif issubclass(object.__class__, hec.data.tx.DataSetTx)            : self.writeTimeSeries(object, storeRule)
		elif issubclass(object.__class__, hec.hecmath.RatingMath)           : self.writeRating(object, storeRule)
		else : raise ValueError, "Don't know how to write a %s" % `object.__class__`
			
        def writeTimeSeries(self, ds, storeRule=None) :
		'''
		Write a time-series supplied as a DataSetTx or TimeSeriesMath object to the database.
		'''
		if issubclass(ds.__class__, hec.hecmath.TimeSeriesMath) :
			_ds = tsm2dtx(ds)
		else :
			_ds = ds
		self._clientLock.acquire()
		self._checkStatus()
		#----------------------------#
		# analyze the DataSetTx data #
		#----------------------------#
		if _ds.getNumberValues() == 0:
			sys.stderr.write("No values supplied.\n")
			self._clientLock.release()
			return
		timeVal = hec.heclib.util.HecTime()
		timeVal.setTimeInMillis(_ds.getStartTime())
		startTime = timeVal.dateAndTime(4).replace(",", "").replace(":", "")
		timeVal.setTimeInMillis(_ds.getEndTime())
		endTime = timeVal.dateAndTime(4).replace(",", "").replace(":", "")
		timeWindow = "%s %s" % (startTime, endTime)
		tsId = _ds.getDescriptionTx().toString()
		#-----------------------#
		# handle the store rule #
		#-----------------------#
		if storeRule :
			storeRuleToUse = getStoreRuleNumber(storeRule)
		else :
			storeRuleToUse = self._storeRule
		#-----------------------------------------------#
		# create a the structures and submit to the DBI #
		#-----------------------------------------------#
		action = wcds.lang.Action(_ds, 
			_ds.getDescriptionTx(), 
			wcds.lang.Action.STORE, 
			storeRuleToUse, 
			"DBI", 
			self._name)
		self.transactWithDBI(action)
		self._clientLock.release()
		
        def writeRating(self, rating, storeRule=None) :
		'''
		Write a rating supplied as a RatingTimeSeries or RatingMath object to the database.
		'''
		if issubclass(rating.__class__, hec.hecmath.RatingMath) :
			_rating = srm2srs(rating)
		else :
			_rating = rating
		_rating.convertToSIUnits()			
		self._clientLock.acquire()
		self._checkStatus()
		#-------------------------#
		# analyze the rating data #
		#-------------------------#
		try :
			if not _rating or _rating.length() == 0:
				sys.stderr.write("Empty supplied.\n")
				self._clientLock.release()
				return
		except :
			sys.stderr.write("Error analyzing rating.\n")
			self._clientLock.release()
			return
		timeVal = hec.heclib.util.HecTime()
		timeVal.setTimeInMillis(_rating.getEffectiveTimeLong(0))
		startTime = timeVal.dateAndTime(4).replace(",", "").replace(":", "")
		timeVal.setTimeInMillis(_rating.getEffectiveTimeLong(_rating.length()-1))
		endTime = timeVal.dateAndTime(4).replace(",", "").replace(":", "")
		timeWindow = "%s %s" % (startTime, endTime)
		ratingId = _rating.toString()
		#-----------------------#
		# handle the store rule #
		#-----------------------#
		if storeRule :
			storeRuleToUse = getStoreRuleNumber(storeRule)
		else :
			storeRuleToUse = self._storeRule
		#-----------------------------------------------#
		# create a the structures and submit to the DBI #
		#-----------------------------------------------#
		actionSpecifier = hec.data.rating.ActionSpecifier(_rating, storeRuleToUse)
		action = actionSpecifier.getAction("DBI", self._name)
		action.setDescription("Storing %s , TW=[%s]" % (ratingId, timeWindow))
		self.transactWithDBI(action)
		self._clientLock.release()
		
        def updateRatingKey(self, ratingID, newRatingID, interpolateBehavior, underflowBehavior, overflowBehavior):
		'''
		Update the info about a rating in the database.
		'''
		self._clientLock.acquire()
		self._checkStatus()
		#-----------------------------------------------#
		# create a the structures and submit to the DBI #
		#-----------------------------------------------#
		actionSpecifier = hec.data.rating.ActionSpecifier(
			ratingID,
			newRatingID,
			interpolateBehavior,
			underflowBehavior,
			overflowBehavior)
		action = actionSpecifier.getAction("DBI", self._name)
		action.setDescription(
			"Updating %s to %s (%s, %s, %s)" % 
			(ratingID, newRatingID, interpolateBehavior, underflowBehavior, overflowBehavior))
		self.transactWithDBI(action)
		self._clientLock.release()
		
	def deleteRating(self, ratingID):
		'''
		Delete a rating (key and all data) from database.
		'''
		self._clientLock.acquire()
		self._checkStatus()
		#-----------------------------------------------#
		# create a the structures and submit to the DBI #
		#-----------------------------------------------#
		actionSpecifier = hec.data.rating.ActionSpecifier(ratingID, 1, 1)
		action = actionSpecifier.getAction("DBI", self._name)
		action.setDescription("Deleting rating %s (key and all data)" % ratingID)
		self.transactWithDBI(action)
		self._clientLock.release()
			
		
	def transactWithDBI(self, action) :
		'''
		Sends action to DBI and waits for callback
		'''
		#----------------------------#
		# send the action to the DBI #
		#----------------------------#
		actionCallback = hec.msgsystem.ObservableActionCallbackImpl()
		actionCallback.setAction(action)
		actionCallback.addObserver(self)
		remoteWrapper = hec.server.RemoteWrapper(actionCallback)
		remoteWrapper.export()
		self._updated = 0
		self._returnValue = None
		self._transactionTime = None
		self._startAction = time.time()
		self._dbiWrapper.sendToDbi(actionCallback)
		#------------------------------------------------#
		# wait for and process the callback from the DBI #
		#------------------------------------------------#
		while not self._updated :
			if time.time() - self._startAction > self._timeout : break
			time.sleep(0.005)
		if not self._updated :
			self._clientLock.release()
			raise IOError, "The DBI did not respond within the timeout of %d seconds." % self._timeout
		
        def update(self, observable, arg) :
		'''
		Method the DBI calls to let us know it's done.
		'''
		self._callbackLock.acquire()
		self._endAction = time.time()
		self._transactionTime = self._endAction - self._startAction
		if issubclass(observable.__class__, hec.msgsystem.ObservableActionCallbackImpl) :
			actionCallback = observable
			self._dbiWrapper.removeActionCallback(actionCallback)
			actionCallback.deleteObserver(self)
			action = actionCallback.getAction()
			if action and issubclass(action.__class__, wcds.lang.Action) :
				self._returnValue = action.getReturnValue()
				self._data = None
				data = action.getData()
				if data :
					if issubclass(data.__class__, hec.data.tx.DataSetTx) :
						self._data = data
					elif issubclass(data.__class__, hec.data.rating.RatingTimeSeries) :
						self._data = data
		self._updated = 1
		self._callbackLock.release()
		
	def close(self) :
		'''
		Release any resources and make this object unusable.
                '''
		self._clientLock.acquire()
		self._initValues()
		self._clientLock.release()
			
			
##############################################################################

TSCatalogMagicString = "~@~@~@~@~@~@TSCatalog@~@~@~@~@~@~"
RatingCatalogMagicString = "+-+-+-+-+-+-RatingCatalog-+-+-+-+-+-+"

def openDbi(url=None, startTime=None, endTime=None) :
	'''
	Return a new DbiAccess object.
	'''
	return DbiAccess(url, startTime, endTime)	
	
def readTSCatalog(fileName, maxAge=0) :
	'''
	Retrieve a time-series catalog from disk.
	'''
	#---------------#
	# open the file #
	#---------------#
	try :    
		file = open(fileName,  "r")
	except : 
		return None		
	#--------------------------------------#
	# verify that it's a catalog save file #
	#--------------------------------------#
	line = file.readline().strip()
	if line != TSCatalogMagicString :
		file.close()
		return None
	#--------------------------------------------------#
	# handle the max age (in hours) if one is supplied #
	#--------------------------------------------------#
	line = file.readline().strip()
	if maxAge :
		try :
			age = (time.time() - float(line)) / 3600.
			if age > maxAge :
				sys.stderr.write("Catalog save file %s is out of date.\n" % fileName)
				file.close()
				return None
		except :
			file.close()
			return None
	#-------------------------------------------------------#
	# read the remainder of the file and create the catalog #
	#-------------------------------------------------------#
	lines = file.readlines()
	file.close()
	boundaries = []
	for i in range(len(lines)) :
		lines[i] = lines[i].strip()
		if lines[i] == TSCatalogMagicString :
			boundaries.append(i)
			if len(boundaries) == 8 : break
	if len(boundaries) == 8 :
		tsids = lines[:boundaries[0]]
		parts = [ \
			lines[boundaries[0]+1:boundaries[1]],
			lines[boundaries[1]+1:boundaries[2]],
			lines[boundaries[2]+1:boundaries[3]],
			lines[boundaries[3]+1:boundaries[4]],
			lines[boundaries[4]+1:boundaries[5]],
			lines[boundaries[5]+1:boundaries[6]],
			lines[boundaries[6]+1:boundaries[7]],
			lines[boundaries[7]+1:]
		]
		cat = TSCatalog(tsids, parts)
	else :
		cat = TSCatalog(lines)
	return cat
	
def readRatingCatalog(fileName, maxAge=0) :
	'''
	Retrieve a rating catalog from disk.
	'''
	#---------------#
	# open the file #
	#---------------#
	try :    
		file = open(fileName,  "r")
	except : 
		return None		
	#--------------------------------------#
	# verify that it's a catalog save file #
	#--------------------------------------#
	line = file.readline().strip()
	if line != RatingCatalogMagicString :
		file.close()
		return None
	#--------------------------------------------------#
	# handle the max age (in hours) if one is supplied #
	#--------------------------------------------------#
	line = file.readline().strip()
	if maxAge :
		try :
			age = (time.time() - float(line)) / 3600.
			if age > maxAge :
				sys.stderr.write("Catalog save file %s is out of date.\n" % fileName)
				file.close()
				return None
		except :
			file.close()
			return None
	#-------------------------------------------------------#
	# read the remainder of the file and create the catalog #
	#-------------------------------------------------------#
	lines = file.readlines()
	file.close()
	boundaries = []
	for i in range(len(lines)) :
		lines[i] = lines[i].strip()
		if lines[i] == RatingCatalogMagicString :
			boundaries.append(i)
			if len(boundaries) == 10 : break
	if len(boundaries) == 10 :
		ratingIds = lines[:boundaries[0]]
		parts = [ \
			lines[boundaries[0]+1:boundaries[1]],
			lines[boundaries[1]+1:boundaries[2]],
			lines[boundaries[2]+1:boundaries[3]],
			lines[boundaries[3]+1:boundaries[4]],
			lines[boundaries[4]+1:boundaries[5]],
			lines[boundaries[5]+1:boundaries[6]],
			lines[boundaries[6]+1:boundaries[7]],
			lines[boundaries[7]+1:boundaries[8]],
			lines[boundaries[8]+1:boundaries[9]],
			lines[boundaries[9]+1:]
		]
		cat = RatingCatalog(ratingIds, parts)
	else :
		cat = RatingCatalog(lines)
	return cat
	
def getStoreRuleNumber(storeRule) :
	'''
	Take an int or a string and pass back a validated rule number.
	'''
	if type(storeRule) == type("") :
		number = hec.lang.Const.getRuleNumber(storeRule)
		if number == hec.lang.Const.UNDEFINED_INT :
			raise ValueError, "Invalid store rule"
		return number	
	elif type(storeRule) == type(0):
		if storeRule in hec.lang.Const.RULE_NUMBERS :
			return storeRule
		else :
			raise ValueError, "Invalid store rule"
	else :
		raise TypeError, "Store rule must be an integer or string."
		

def dtx2tsm(dtx) :
	'''
	Convert a DataSetTx object to a TimeSeriesMath object.
	'''
	if not dtx : return None
	tsc = dtx.getTimeSeriesContainer()
	if tsc.interval > 0 : tsc.interval /= 60
	for i in range(len(tsc.values)) :
		if not hec.lang.Const.isPracticallyUndefined(tsc.values[i]) :
			values = list(tsc.values)[i:]
			times = list(tsc.times)[i:]
			break
	else :
		values = []
		times = []
	for i in range(len(values) - 1, -1, -1) :
		if not hec.lang.Const.isPracticallyUndefined(values[i]) :
			values = values[:i+1]
			times = times[:i+1]
			break
	else :
		values = []
		times = []
		
	tsc.numberValues = len(values)
	tsc.values = values
	tsc.times = times
	return hec.hecmath.HecMath.createInstance(tsc)

def tsm2dtx(tsm) :
	'''
	Convert a TimeSeriesMath object to a DataSetTx object.
	'''
	if not tsm : return None
	return hec.data.tx.DataSetTx(tsm.getData())
	
def srs2srm(srs) :
	'''
	Convert a StreamRatingSeries object to a StreamRatingMath object.
	'''
	if not srs : return None
	return hec.hecmath.HecMath.createInstance(srs.getData())
	
def srm2srs(srm) :
	'''
	Convert a StreamRatingMath object to a StreamRatingSeries object.
	'''
	if not srm : return None
	return hec.data.rating.StreamRatingSeries.createStreamRatingSeriesFromContainer(srm.getData())

def getVersion() :
	'''
	Returns the version of this module in the form "X.XX ddMMMyyyy"
	'''
	return "1.11 11Sep2004"

